// VideoMemoryAllocator class implementation
#include "pch.h"

#include "VideoMemoryAllocator.h"
#include "util.h"

// 1 GiB
#define HEAP_SIZE_LARGE 1024ull * 1024 * 1024
// 50 MiB
#define DEFAULT_ALLOCATION_AMOUNT 50ull * 1024 * 1024

namespace nf::client::render {
    VideoMemoryAllocator::VideoMemoryAllocator(const VkDevice& device, const VkPhysicalDevice& physicalDevice)
        : m_device(device)
        , m_physicalDevice(physicalDevice)
        , m_blocks()
    {}

    void VideoMemoryAllocator::allocateForBuffer(VkBuffer buffer, VkMemoryPropertyFlags memoryPropertyFlags, VkDeviceMemory& outDeviceMemory, VkDeviceSize& outOffset) {
        VkMemoryRequirements memoryRequirements = {};
        vkGetBufferMemoryRequirements(m_device, buffer, &memoryRequirements);

        VkMemoryAllocateInfo allocateInfo = { VK_STRUCTURE_TYPE_MEMORY_ALLOCATE_INFO };
        allocateInfo.allocationSize = memoryRequirements.size;
        allocateInfo.memoryTypeIndex = findMemoryType(memoryRequirements.memoryTypeBits, memoryPropertyFlags);

        if (allocateInfo.memoryTypeIndex == -1)
            NFError("Could not find suitable memory type.");

        allocateFromBlock(allocateInfo, memoryRequirements.alignment, outDeviceMemory, outOffset);
    }

    void VideoMemoryAllocator::deallocate(VkDeviceMemory memoryBlock, VkDeviceSize offset) {
        const auto blockIterator = std::find_if(m_blocks.begin(), m_blocks.end(), [&](const auto& currentBlock) {
            if (currentBlock.memory == memoryBlock)
                return true;
            else return false;
        });

        if (blockIterator == m_blocks.end())
            NFError("Could not find video memory block to deallocate from.");

        MemoryBlock& block = *blockIterator;

        const auto allocationIterator = block.allocations.find(offset);

        if (allocationIterator == block.allocations.end())
            NFError("Could not find video memory allocation to deallocate.");

        block.allocations.erase(allocationIterator);

        // If block is now empty, free it
        if (block.allocations.empty()) {
            vkFreeMemory(m_device, block.memory, nullptr);
            m_blocks.erase(blockIterator);
        }
    }

    uint32_t VideoMemoryAllocator::findMemoryType(uint32_t typeFilter, VkMemoryPropertyFlags desiredPropertyFlags) {
        VkPhysicalDeviceMemoryProperties memoryProperties = {};
        vkGetPhysicalDeviceMemoryProperties(m_physicalDevice, &memoryProperties);

        for (int i = 0; i < memoryProperties.memoryTypeCount; i++)
            if (typeFilter & (1 << i) && (memoryProperties.memoryTypes[i].propertyFlags & desiredPropertyFlags) == desiredPropertyFlags)
                return i;

        return -1;
    }

    void VideoMemoryAllocator::allocateFromBlock(VkMemoryAllocateInfo& allocateInfo, uint32_t alignment, VkDeviceMemory& outDeviceMemory, VkDeviceSize& outOffset) {
        while (true) {
            for (auto& block : m_blocks) {
                if (block.memoryTypeIndex == allocateInfo.memoryTypeIndex) {
                    VkDeviceSize lastOffset = 0, lastSize = 0;
                    // Check between existing allocations
                    for (const auto& [currentOffset, currentSize] : block.allocations) {
                        // Check for alignment
                        VkDeviceSize checkOffset = lastOffset + lastSize, remainder = checkOffset % alignment;
                        checkOffset = remainder ? checkOffset + (alignment - remainder) : checkOffset;
                        // Check for enough space
                        if (currentOffset - lastOffset - lastSize >= allocateInfo.allocationSize) {
                            outDeviceMemory = block.memory;
                            outOffset = checkOffset;
                            block.allocations[checkOffset] = allocateInfo.allocationSize;
                            return;
                        }

                        lastOffset = currentOffset, lastSize = currentSize;
                    }

                    // Check end
                    VkDeviceSize checkOffset = lastOffset + lastSize, remainder = checkOffset % alignment;
                    checkOffset = remainder ? checkOffset + (alignment - remainder) : checkOffset;
                    if ((block.size - 1) - lastOffset - lastSize >= allocateInfo.allocationSize) {
                        outDeviceMemory = block.memory;
                        outOffset = checkOffset;
                        block.allocations[checkOffset] = allocateInfo.allocationSize;
                        return;
                    }

                    // If not enough space, continue
                    continue;
                }
            }

            // No suitable block found, allocate one
            VkMemoryAllocateInfo newBlockAllocateInfo = { VK_STRUCTURE_TYPE_MEMORY_ALLOCATE_INFO };
            newBlockAllocateInfo.memoryTypeIndex = allocateInfo.memoryTypeIndex;
            newBlockAllocateInfo.allocationSize = calculateBlockSize(allocateInfo.memoryTypeIndex);
            allocateBlock(newBlockAllocateInfo);
        }
    }

    size_t VideoMemoryAllocator::calculateBlockSize(uint32_t memoryTypeIndex) const {
        VkPhysicalDeviceMemoryProperties memoryProperties = {};
        vkGetPhysicalDeviceMemoryProperties(m_physicalDevice, &memoryProperties);
        size_t heapSize = memoryProperties.memoryHeaps[memoryProperties.memoryTypes[memoryTypeIndex].heapIndex].size;

        return heapSize > HEAP_SIZE_LARGE ? DEFAULT_ALLOCATION_AMOUNT : (heapSize / 8);
    }

    void VideoMemoryAllocator::allocateBlock(VkMemoryAllocateInfo& allocateInfo) {
        VkDeviceMemory memoryBlock = nullptr;

        if (vkAllocateMemory(m_device, &allocateInfo, nullptr, &memoryBlock) != VK_SUCCESS)
            NFError("Could not allocate video memory.");

        m_blocks.emplace_back(memoryBlock, allocateInfo.memoryTypeIndex, allocateInfo.allocationSize);
    }
}